"""
P1 Data API exceptions.

Import as

import p1_data_client_python.exceptions as p1_exc
"""


class UnauthorizedException(Exception):
    pass


class BadMetaDataTypeException(Exception):
    pass


class ParseResponseException(Exception):
    pass


class TestTokenNotFound(Exception):
    pass



from typing import List
import unittest.mock as mock

import helpers.unit_test as hut
import p1_data_client_python.client as p1_data
import p1_data_client_python.exceptions as p1_exc
import pandas as pd

TOKEN_ENV_NAME = "P1_API_TOKEN"
EXAMPLE_METADATA_TYPE = "COMMODITIES"
BAD_METADATA_TYPE = "Stranger Things"
SEARCH_ROW_EXAMPLE = {
    "name": "qweqwe",
    "commodity": "qweqweqwe",
    "payload_id": "asdasd",
    "business_category": "asdasd",
    "country": "asdasd",
    "frequency": "asdasd",
    "unit": "zxc",
    "start_date": "zxcff",
}


class SearchOnePageGoodResponse:
    status_code = 200

    @staticmethod
    def json() -> dict:
        return {
            "scroll_id": "qweqwe",
            "total_count": 5555,
            "rows": [SEARCH_ROW_EXAMPLE for _ in range(5555)],
        }


class PayloadGoodResponseMock:
    status_code = 200

    @staticmethod
    def json() -> dict:
        return {
            "payload_data": [
                {
                    "original_period": "010101",
                    "original_value": "1.33",
                    "period": "01-01-01",
                    "value": 1.33,
                },
                {
                    "original_period": "020202",
                    "original_value": "4.33",
                    "period": "02-02-02",
                    "value": 4.33,
                },
            ]
        }


class MessyResponseMock:
    status_code = 200

    @staticmethod
    def json() -> dict:
        return {"message": "strange_message"}


class MetaDataGoodResponseMock:
    status_code = 200

    @staticmethod
    def json() -> dict:
        return {"data": [{"name": "Metadata1"}, {"name": "Metadata2"}]}


class TestPythonClientMock(hut.TestCase):
    def setUp(self) -> None:
        self.client = p1_data.Client(token="goo token")
        super().setUp()

    def test_list_of_metadata(self) -> None:
        self.assertIsInstance(self.client.list_of_metadata, List)

    @mock.patch("requests.Session.request")
    def test_search(self, mock_request):
        # test on UnauthorizedException
        mock_request.return_value = mock.Mock(status_code=401)
        with self.assertRaises(p1_exc.UnauthorizedException):
            self.client.search(text="Price")
        # test on good response
        mock_request.return_value = SearchOnePageGoodResponse()
        self.client.search(text="Price")
        for page in self.client.search_pages(pages_limit=2):
            self.assertIsInstance(page, pd.DataFrame)

    @mock.patch("requests.Session.request")
    def test_payload(self, mock_request) -> None:
        payload_id = "some very good get_payload ID"
        # test on UnauthorizedException
        mock_request.return_value = mock.Mock(status_code=401)
        with self.assertRaises(p1_exc.UnauthorizedException):
            self.client.get_payload(payload_id)
        # test on good response
        mock_request.return_value = PayloadGoodResponseMock()
        self.assertIsInstance(self.client.get_payload(payload_id), pd.DataFrame)
        # test on ParseResponseException
        mock_request.return_value = MessyResponseMock()
        with self.assertRaises(p1_exc.ParseResponseException):
            self.client.get_payload(payload_id)

    @mock.patch("requests.Session.request")
    def test_get_metadata_type_mock(self, mock_request) -> None:
        # test on UnauthorizedException
        mock_request.return_value = mock.Mock(status_code=401)
        with self.assertRaises(p1_exc.UnauthorizedException):
            self.client.get_metadata_type(EXAMPLE_METADATA_TYPE)
        # test on good response
        mock_request.return_value = MetaDataGoodResponseMock
        self.assertIsInstance(
            self.client.get_metadata_type(EXAMPLE_METADATA_TYPE), pd.DataFrame
        )
        # test on ParseResponseException
        mock_request.return_value = MessyResponseMock
        with self.assertRaises(p1_exc.ParseResponseException):
            self.client.get_metadata_type(EXAMPLE_METADATA_TYPE)



"""
The core of P1 Data REST API wrapper

Copyright 2020 by Particle.One Inc.
All rights reserved.
This file is part of the ParticleOne and is released under the "MIT". Please
see the license.txt file that should have been included as part of this
package.


Import as

import p1_data_client_python.client as p1_data
"""

import json
from typing import Any, Dict, List

import p1_data_client_python.exceptions as p1_exc
import pandas as pd
import requests
import requests.adapters as rq_adapt
import requests.packages.urllib3.util.retry as rq_retry

DEFAULT_BASE_URL = "https://data.particle.one"


class Client:
    """
    Base class for p1 data REST API operating.
    """

    SEARCH_CHUNK_SIZE = 1000
    API_ROUTES = {
        "AUTH": "/auth-token/",
        "SEARCH": "/data-api/v1/search/",
        "SEARCH_SCROLL": "/data-api/v1/search-scroll/",
        "PAYLOAD": "/data-api/v1/payload/",
    }
    METADATA_ROUTES = {
        "COMMODITIES": "/data-api/v1/commodities/",
        "BUSINESS-CATEGORIES": "/data-api/v1/business-categories/",
        "COUNTRIES": "/data-api/v1/countries/",
        "FREQUENCIES": "/data-api/v1/frequencies/",
    }

    def __init__(
        self,
        token: str,
        base_url: str = DEFAULT_BASE_URL,
        use_retries: bool = True,
        retries_number: int = 5,
        backoff_factor: float = 0.3,
    ):
        """
        Pass arguments and gets authenticated in the system.
        :param base_url: REST API Server url.
        :param token: Your token for access to the system.
        """
        self.base_url = base_url.rstrip("//")
        self.token = token
        self.use_retries = use_retries
        self.retries_number = retries_number
        self.backoff_factor = backoff_factor
        self._scroll_id = ""
        self.status_forcelist = (500, 502, 504)
        self._last_search_parameters = None
        self.session = self._get_session()
        self.headers = {
            "Authorization": "Token " + self.token,
            "Content-Type": "application/json",
        }

    def _get_session(self) -> requests.Session:
        """
        Initialize and return a session allows make retry
        when some errors will raised.
        """
        session = requests.Session()
        retry = rq_retry.Retry(
            total=self.retries_number,
            read=self.retries_number,
            connect=self.retries_number,
            backoff_factor=self.backoff_factor,
            status_forcelist=self.status_forcelist,
        )
        adapter = rq_adapt.HTTPAdapter(max_retries=retry)
        session.mount("https://", adapter)
        return session

    def _make_request(self, *args, **kwargs) -> requests.Response:
        """
        Single entry point for any request to the REST API.
        :return: requests Response.
        """
        response = self.session.request(*args, **kwargs)
        # Throw exception, if token is not valid.
        if response.status_code == 401:
            raise p1_exc.UnauthorizedException(response.text)
        return response

    @property
    def list_of_metadata(self) -> List[str]:
        """
        Retrieve list of metadata keys from METADATA_ROUTES
        """
        return list(self.METADATA_ROUTES.keys())

    def search_pages(self, pages_limit: int = 100) -> pd.DataFrame:
        """
        Get generator which scrolling down through pages
        until end or limit will reached.
        :param pages_limit: Max paged for scrolling.
        """
        current_page_number = 1
        row_count = self._last_total_count

        if self._scroll_id:
            while True:
                if row_count > 0 and current_page_number < pages_limit:
                    payloads = self.search_scroll()
                else:
                    break
                row_count = len(payloads["rows"])
                self._scroll_id = payloads["scroll_id"]
                yield pd.DataFrame(payloads["rows"])
                current_page_number += 1

    def search_scroll(self) -> Dict[Any, Any]:
        """
        Get next chunk(page) of payloads by given scroll_id
        """
        response = self._make_request(
            "GET",
            self.base_url + self.API_ROUTES["SEARCH_SCROLL"],
            headers=self.headers,
            params={"scroll_id": self._scroll_id},
        )
        try:
            next_page = response.json()
        except json.JSONDecodeError as e:
            raise json.JSONDecodeError(
                msg=f"Can't decode response: "
                f"{response.text} as JSON, {repr(e)}",
                doc=e.doc,
                pos=e.pos,
            )
        return next_page

    def _parse_search(self, response: requests.Response) -> pd.DataFrame:
        """
        Parse search response and return pandas Dataframe
        """
        payloads = response.json()
        self._scroll_id = payloads["scroll_id"]
        self._last_total_count = payloads["total_count"]

        return pd.DataFrame(payloads["rows"])

    def search(self, **search_payload) -> pd.DataFrame:
        """
        Get payloads IDs by given search conditions.
        :param text:
        :param commodity:
        :param business_category:
        :param country:
        :param frequency:
        :return: A first page (chunk of a data) that represent.
        pandas dataframe: pd.DataFrame
        """
        # Store search a last search conditions.
        self._last_search_parameters = search_payload
        response = self._make_request(
            "POST",
            self.base_url + self.API_ROUTES["SEARCH"],
            headers=self.headers,
            data=json.dumps(self._last_search_parameters),
        )
        # Parse response to a pandas dataframe, check for errors.
        try:
            search_dataframe = self._parse_search(response)
        except (KeyError, json.JSONDecodeError) as e:
            raise p1_exc.ParseResponseException(
                "Can't transform server response to a pandas Dataframe"
            ) from e
        return search_dataframe

    @staticmethod
    def _parse_payload(response: requests.Response) -> pd.DataFrame:
        """
        Parse payload response and return pandas Dataframe.
        """
        payload_response = response.json()
        return pd.DataFrame(payload_response["payload_data"])

    def get_payload(self, payload_id: str) -> pd.DataFrame:
        """
        Get time series data by payload_id.
        :param payload_id: ID of payload from search method.
        """
        response = self._make_request(
            "GET",
            self.base_url + self.API_ROUTES["PAYLOAD"],
            headers=self.headers,
            params={"payload_id": payload_id},
        )
        # Parse response to a pandas dataframe, check for errors.
        try:
            payload_dataframe = self._parse_payload(response)
        except (KeyError, json.JSONDecodeError) as e:
            raise p1_exc.ParseResponseException(
                "Can't transform server response to a pandas Dataframe"
            ) from e
        return payload_dataframe

    @staticmethod
    def _parse_metadata_type(
        metadata_type: str, response: requests.Response
    ) -> pd.DataFrame:
        """
        Parse metadata_type response and return pandas Dataframe
        """
        metadata_response = response.json()
        metadata_list = [row["name"] for row in metadata_response["data"]]
        return pd.DataFrame(metadata_list, columns=[metadata_type])

    def get_metadata_type(self, metadata_type: str) -> pd.DataFrame:
        """
        Get list of values for any metadata type.
        All types are listed in METADATA_ROUTES.
        :return: pandas Dataframe with metadata type values on-board
        """
        # Check if metadata_type in the allowed list.
        try:
            metadata_type_path = self.METADATA_ROUTES[metadata_type]
        except KeyError as e:
            raise p1_exc.BadMetaDataTypeException(
                f"{metadata_type} metadata "
                f"type is not supported in the client"
            ) from e
        # make request
        response = self._make_request(
            "GET", self.base_url + metadata_type_path, headers=self.headers
        )
        # Parse response to a pandas dataframe, check for errors.
        try:
            metadata_type_dataframe = self._parse_metadata_type(
                metadata_type, response
            )
        except (KeyError, json.JSONDecodeError) as e:
            raise p1_exc.ParseResponseException(
                "Can't transform server response to a pandas Dataframe"
            ) from e
        return metadata_type_dataframe



import os

import helpers.unit_test as hut
import p1_data_client_python.client as p1_data
import p1_data_client_python.exceptions as p1_exc
import pandas as pd
import pytest

# TODO: Remove alpha url before production usage
TOKEN_ENV_NAME = "P1_API_TOKEN"
TEST_TIMEOUT_URL = "https://httpstat.us/200?sleep=50000"
BAD_TOKEN = "d238f0111111"
EXAMPLE_METADATA_TYPE = "COMMODITIES"
BAD_METADATA_TYPE = "Stranger Things"


class TestPythonClient(hut.TestCase):
    def set_env_token(self):
        self.env_token = os.getenv(TOKEN_ENV_NAME)
        if not self.env_token:
            raise p1_exc.TestTokenNotFound(
                f"Token not found in the environment variable {TOKEN_ENV_NAME}"
            )

    def setUp(self) -> None:
        self.set_env_token()
        self.client = p1_data.Client(token=self.env_token)
        super().setUp()

    def test_search(self) -> None:
        # test on good response
        self.client.search(text="Price")
        for page in self.client.search_pages(pages_limit=2):
            self.assertIsInstance(page, pd.DataFrame)
        # test on UnauthorizedException
        with self.assertRaises(p1_exc.UnauthorizedException):
            self.client.token = BAD_TOKEN
            self.client.headers["Authorization"] = "Token " + self.client.token
            self.client.search(text="Price")

    def test_search_scroll(self) -> None:
        self.client.search(text="Price")
        self.assertIsInstance(self.client.search_scroll(), dict)

    def test_payload(self) -> None:
        payloads = self.client.search(text="Price")
        payload_time_series_data = self.client.get_payload(
            payloads.iloc[0].payload_id
        )
        self.assertIsInstance(payload_time_series_data, pd.DataFrame)

    def test_get_metadata_type(self) -> None:
        for metadata_type in self.client.METADATA_ROUTES:
            isinstance(self.client.get_metadata_type(metadata_type), pd.DataFrame)
        with self.assertRaises(p1_exc.UnauthorizedException):
            self.client.token = BAD_TOKEN
            self.client.headers["Authorization"] = "Token " + self.client.token
            self.client.get_metadata_type(EXAMPLE_METADATA_TYPE)
        with pytest.raises(p1_exc.BadMetaDataTypeException):
            self.client.get_metadata_type(BAD_METADATA_TYPE)



import sys

if sys.version_info[:2] < (3, 7):
    raise ImportError(
        """
    This version of p1_data_client_python don't support python versions less than 3.7
    """
    )

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

INSTALL_REQUIRES = ["pandas>=1.0.0", "requests>=2.18.0"]
TEST_REQUIRES = ["pytest>=5.0.0"]
PACKAGES = [
    "p1_data_client_python",
]

project_urls = {
  'Site': 'https://particle.one/',
  'API registration': 'https://particle.one/api'
}

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    version="1.0.2",
    name="p1_data_client_python",
    description="Package for P1 Data API access",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords=["p1_data_client_python", "API",
              "data", "financial", "economic",
              "particle", "particleone", "particle.one"],
    author="GP Saggese, Paul Smith",
    author_email="gp@particle.one, paul@particle.one",
    maintainer="",
    maintainer_email="",
    url="https://github.com/ParticleDev/p1_data_client_python",
    license="MIT",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    install_requires=INSTALL_REQUIRES,
    tests_require=TEST_REQUIRES,
    python_requires=">= 3.7",
    test_suite="pytest",
    packages=PACKAGES,
    project_urls=project_urls
)



import helpers.git as git
import helpers.unit_test as hut


class TestClientJupyterNotebook(hut.TestCase):
    def test_notebook1(self) -> None:
        file_name = git.find_file_in_git_tree("p1_data_client_example.ipynb")
        scratch_dir = self.get_scratch_space()
        hut.run_notebook(file_name, scratch_dir)
